/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "condor_common.h"
#include "condor_nfs.h"
#include "condor_debug.h"
#include "condor_syscall_mode.h"
#include "std_univ_io.h"
#include "condor_sys.h"
#include "condor_mmap.h"
#include "condor_error.h"
#include "startup.h"
#include "syscall_param_sizes.h"
#include "../condor_ckpt/signals_control.h"
#include "condor_open.h"

#ifdef LINK_PVM
// ClassAds needed for start_work_request used by PVM starter/shadow
#include "condor_classad.h"
#endif

/* get rid of the mapping of access to access_euid for the syscall library */
#ifdef access
#	undef access
#endif

int CurrentSysCall;

extern StdUnivSock *syscall_sock;

/* Ahhh, like drinking beer by the pool side. */
extern "C" {
#if defined( SYS_open )
int  
REMOTE_CONDOR_open(char *  path , open_flags_t   flags , long   lastarg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_open;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->code(lastarg) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_close )
int  
REMOTE_CONDOR_close(int   fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_close;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_unlink )
int  
REMOTE_CONDOR_unlink(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_unlink;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_rename(char *  from , char *  to)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_rename;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(to) );
	assert( syscall_sock->code(from) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_read )
ssize_t  
REMOTE_CONDOR_read(int   fd , void *  buf , size_t   len)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_read;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_write )
ssize_t  
REMOTE_CONDOR_write(int   fd , void *  buf , size_t   len)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_write;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code_bytes_bool(buf, len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_lseek )
off_t  
REMOTE_CONDOR_lseek(int   fd , off_t   offset , int   whence)
{
	int	scm;
	off_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lseek;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(offset) );
	assert( syscall_sock->code(whence) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

off64_t  
REMOTE_CONDOR_llseek(int   fd , off64_t   offset , int   whence)
{
	int	scm;
	off64_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lseek;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(offset) );
	assert( syscall_sock->code(whence) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_dup )
int  
REMOTE_CONDOR_dup(int   old_fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_dup;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(old_fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_dup2 )
int  
REMOTE_CONDOR_dup2(int   old_fd , int   new_fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_dup2;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(old_fd) );
	assert( syscall_sock->code(new_fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_fcntl )
int  
REMOTE_CONDOR_fcntl(int   fd , fcntl_cmd_t   cmd , long   lastarg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fcntl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->code(lastarg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_ioctl )
int  
REMOTE_CONDOR_ioctl(int   fd , unsigned long   cmd , long   lastarg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_ioctl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->code(lastarg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_pipe(int  *  fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_pipe;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_fsync(int   fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fsync;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_sync()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_sync;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_ftruncate(int   fd , off_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_ftruncate;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_ftruncate64 )
int  
REMOTE_CONDOR_ftruncate64(int   fd , off64_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_ftruncate;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_fchdir )
int  
REMOTE_CONDOR_fchdir(int   fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fchdir;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_chdir(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_chdir;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_poll )
int  
REMOTE_CONDOR_poll(struct pollfd *  fds , unsigned long   nfds , int   timeout)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_poll;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(nfds) );
	assert( syscall_sock->code(timeout) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_select )
int  
REMOTE_CONDOR_select(int   nfds , fd_set *  readfds , fd_set *  writefds , fd_set *  exceptfds , struct timeval *  timeout)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_select;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(nfds) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_socket )
int  
REMOTE_CONDOR_socket(int   domain , int   type , int   protocol)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_socket;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(domain) );
	assert( syscall_sock->code(type) );
	assert( syscall_sock->code(protocol) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_newfstatat )
int  
REMOTE_CONDOR___fxstatat64(int   dirfd , char *  pathname , struct stat *  buf , int   flags)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fstatat;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(dirfd) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->code(pathname) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_fstat )
int  
REMOTE_CONDOR_fstat(int   fd , struct stat *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fstat;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_stat )
int  
REMOTE_CONDOR_stat(char *  name , struct stat *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_stat;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_lstat )
int  
REMOTE_CONDOR_lstat(char *  name , struct stat *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lstat;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_fchown )
int  
REMOTE_CONDOR_fchown(int   fd , uid_t   owner , gid_t   group)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fchown;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(owner) );
	assert( syscall_sock->code(group) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_fchmod )
int  
REMOTE_CONDOR_fchmod(int   fd , mode_t   mode)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fchmod;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(mode) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_statfs(char *  path , struct statfs *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_statfs;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_fstatfs )
int  
REMOTE_CONDOR_fstatfs(int   fd , struct statfs *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fstatfs;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_flock )
int  
REMOTE_CONDOR_flock(int   fd , int   op)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_flock;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(op) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_lockf )
int  
REMOTE_CONDOR_lockf(int   fd , int   op , off_t   where)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lockf;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(op) );
	assert( syscall_sock->code(where) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_lockf64 )
int  
REMOTE_CONDOR_lockf64(int   fd , int   op , off64_t   where)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lockf;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(op) );
	assert( syscall_sock->code(where) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

ssize_t  
REMOTE_CONDOR_getdirentries(int   fd , char *  buf , size_t   nbytes , off_t *  basep)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getdirentries;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(nbytes) );
	assert( syscall_sock->code_bytes_bool(basep, EIGHT) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(basep, EIGHT) );
	assert( syscall_sock->code_bytes_bool(buf, ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_getdents )
int  
REMOTE_CONDOR_getdents(int   fd , struct dirent *  buf , size_t   nbytes)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getdents;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(nbytes) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, nbytes) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getdents64 )
int  
REMOTE_CONDOR_getdents64(int   fd , struct dirent64 *  buf , size_t   nbytes)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getdents64;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(nbytes) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, nbytes) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

off_t  
REMOTE_CONDOR_lseekread(int   fd , off_t   offset , int   whence , void *  buf , size_t   len)
{
	int	scm;
	off_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lseekread;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(offset) );
	assert( syscall_sock->code(whence) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, len) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

off_t  
REMOTE_CONDOR_lseekwrite(int   fd , off_t   offset , int   whence , void *  buf , size_t   len)
{
	int	scm;
	off_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_lseekwrite;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(offset) );
	assert( syscall_sock->code(whence) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code_bytes_bool(buf, len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_socketpair )
int  
REMOTE_CONDOR_socketpair(int   domain , int   type , int   protocol , int *  fds)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_socketpair;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(domain) );
	assert( syscall_sock->code(type) );
	assert( syscall_sock->code(protocol) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_accept )
int  
REMOTE_CONDOR_accept(int   fd , struct sockaddr *  addr , socklen_t *  length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_accept;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_bind )
int  
REMOTE_CONDOR_bind(int   fd , struct sockaddr *  addr , socklen_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_bind;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_connect )
int  
REMOTE_CONDOR_connect(int   fd , struct sockaddr *  addr , socklen_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_connect;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getpeername )
int  
REMOTE_CONDOR_getpeername(int   fd , struct sockaddr *  addr , socklen_t *  length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpeername;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getsockname )
int  
REMOTE_CONDOR_getsockname(int   fd , struct sockaddr *  addr , socklen_t *  length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getsockname;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getsockopt )
int  
REMOTE_CONDOR_getsockopt(int   fd , int   level , int   option , void *  value , socklen_t *  length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getsockopt;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(level) );
	assert( syscall_sock->code(option) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_listen )
int  
REMOTE_CONDOR_listen(int   fd , int   count)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_listen;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(count) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_recvfrom )
ssize_t  
REMOTE_CONDOR_recvfrom(int   fd , void *  data , unsigned long   length , int   flags , struct sockaddr *  addr , socklen_t *  addrlength)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_recvfrom;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_recvmsg )
ssize_t  
REMOTE_CONDOR_recvmsg(int   fd , struct msghdr *  msg , int   flags)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_recvmsg;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_sendmsg )
ssize_t  
REMOTE_CONDOR_sendmsg(int   fd , struct msghdr *  msg , int   flags)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_sendmsg;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_sendto )
ssize_t  
REMOTE_CONDOR_sendto(int   fd , void *  data , unsigned long   length , int   flags , struct sockaddr *  addr , socklen_t   addrlength)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_sendto;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->code(addrlength) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setsockopt )
int  
REMOTE_CONDOR_setsockopt(int   fd , int   level , int   option , void *  value , socklen_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setsockopt;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(level) );
	assert( syscall_sock->code(option) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_shutdown )
int  
REMOTE_CONDOR_shutdown(int   fd , int   how)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shutdown;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(how) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_link )
int  
REMOTE_CONDOR_link(char *  p1 , char *  p2)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_link;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(p2) );
	assert( syscall_sock->code(p1) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_access )
int  
REMOTE_CONDOR_access(char *  path , int   access_mode)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_access;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(access_mode) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_mknod )
int  
REMOTE_CONDOR_mknod(char *  path , mode_t   mode , dev_t   device)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_mknod;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(mode) );
	assert( syscall_sock->code(device) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_chmod )
int  
REMOTE_CONDOR_chmod(char *  path , mode_t   mode)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_chmod;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(mode) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_symlink )
int  
REMOTE_CONDOR_symlink(char *  path1 , char *  path2)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_symlink;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(path2) );
	assert( syscall_sock->code(path1) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_readlink )
ssize_t  
REMOTE_CONDOR_readlink(char *  path , char *  buf , size_t   buf_size)
{
	int	scm;
	ssize_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_readlink;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(buf_size) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, buf_size) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_chown )
int  
REMOTE_CONDOR_chown(char *  path , uid_t   owner , gid_t   group)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_chown;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(owner) );
	assert( syscall_sock->code(group) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_truncate )
int  
REMOTE_CONDOR_truncate(char *  path , off_t   length)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_truncate;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(length) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_mkdir )
int  
REMOTE_CONDOR_mkdir(char *  path , mode_t   mode)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_mkdir;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(mode) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_rmdir )
int  
REMOTE_CONDOR_rmdir(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_rmdir;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_utimes )
int  
REMOTE_CONDOR_utimes(char *  path , struct timeval *  times)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_utimes;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(times) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_utime )
int  
REMOTE_CONDOR_utime(char *  path , struct utimbuf *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_utime;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(buf) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_shell(char *  command , int   len)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shell;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code_bytes_bool(command, len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_getwd_special(char * &  path_name)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getwd_special;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(path_name) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

uid_t  
REMOTE_CONDOR_getuid()
{
	int	scm;
	uid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getuid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

gid_t  
REMOTE_CONDOR_getgid()
{
	int	scm;
	gid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getgid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

pid_t  
REMOTE_CONDOR_getppid()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getppid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

pid_t  
REMOTE_CONDOR_getpid()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_getpgid )
pid_t  
REMOTE_CONDOR_getpgid(pid_t   pid)
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpgid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(pid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getpgrp )
pid_t  
REMOTE_CONDOR_getpgrp()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpgrp;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

int  
REMOTE_CONDOR_getlogin(char * &  loginbuf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getlogin;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(loginbuf) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_umask )
mode_t  
REMOTE_CONDOR_umask(mode_t   cmask)
{
	int	scm;
	mode_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_umask;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(cmask) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getpagesize )
int  
REMOTE_CONDOR_getpagesize()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpagesize;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getgroups )
int  
REMOTE_CONDOR_getgroups(int   gidsetsize , gid_t  *  grouplist)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getgroups;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(gidsetsize) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_array(grouplist, gidsetsize) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_gethostname )
int  
REMOTE_CONDOR_gethostname(char *  address , size_t   address_len)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_gethostname;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(address_len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(address, address_len) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getdtablesize )
int  
REMOTE_CONDOR_getdtablesize()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getdtablesize;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setpriority )
int  
REMOTE_CONDOR_setpriority(int   which , int   who , int   priority)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setpriority;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(which) );
	assert( syscall_sock->code(who) );
	assert( syscall_sock->code(priority) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getpriority )
int  
REMOTE_CONDOR_getpriority(int   which , int   who)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getpriority;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(which) );
	assert( syscall_sock->code(who) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_gettimeofday )
int  
REMOTE_CONDOR_gettimeofday(struct timeval *  tp , struct timezone *  tzp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_gettimeofday;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(tp) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_gethostid )
int  
REMOTE_CONDOR_gethostid()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_gethostid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setrlimit )
int  
REMOTE_CONDOR_setrlimit(int   resource , struct rlimit *  rlp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setrlimit;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(resource) );
	assert( syscall_sock->code(rlp) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getrlimit )
int  
REMOTE_CONDOR_getrlimit(int   resource , struct rlimit *  rlp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getrlimit;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(resource) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(rlp) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getdomainname )
int  
REMOTE_CONDOR_getdomainname(char *  name , int   namelen)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getdomainname;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(namelen) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(name, namelen) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_mvalid )
int  
REMOTE_CONDOR_mvalid(char *  addr , size_t   len , int   prot)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_mvalid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code(prot) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(addr, len) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_uswitch )
int  
REMOTE_CONDOR_uswitch(int   cmd , int   value)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_uswitch;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->code(value) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

uid_t  
REMOTE_CONDOR_geteuid()
{
	int	scm;
	uid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_geteuid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

gid_t  
REMOTE_CONDOR_getegid()
{
	int	scm;
	gid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getegid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_syscall_version(char *  version)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_syscall_version;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(version) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_ckpt_platform(char *  platform , int   len)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_ckpt_platform;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code_bytes_bool(platform, len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_fs_domain(char *  domain)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_fs_domain;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(domain) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_uid_domain(char *  domain)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_uid_domain;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(domain) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_universe(int *  univ_id)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_universe;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(univ_id) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_username(char *  uname)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_username;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(uname) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_extern_name(char *  path , char *  buf , int   bufsize)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_extern_name;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(bufsize) );
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code_bytes_bool(buf, bufsize) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_free_fs_blocks(char *  pathname)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_free_fs_blocks;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(pathname) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_reallyexit(int *  status , struct rusage *  use_p)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_reallyexit;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(use_p) );
	assert( syscall_sock->code(status) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_subproc_status(int   id , int *  status , struct rusage *  use_p)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_subproc_status;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(id) );
	assert( syscall_sock->code(use_p) );
	assert( syscall_sock->code(status) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_image_size(int   kbytes)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_image_size;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(kbytes) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_send_rusage(struct rusage *  use_p)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_send_rusage;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(use_p) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_getrusage(int   who , struct rusage *  use_p)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getrusage;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(who) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(use_p) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_startup_info_request(STARTUP_INFO *  s)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_startup_info_request;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(s) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_report_error(char *  msg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_report_error;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(msg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_report_file_info(char *  kind , char *  name , int   open_count , int   read_count , int   write_count , int   seek_count , int   read_bytes , int   write_bytes)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_report_file_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(open_count) );
	assert( syscall_sock->code(read_count) );
	assert( syscall_sock->code(write_count) );
	assert( syscall_sock->code(seek_count) );
	assert( syscall_sock->code(read_bytes) );
	assert( syscall_sock->code(write_bytes) );
	assert( syscall_sock->code(name) );
	assert( syscall_sock->code(kind) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_report_file_info_new(char *  name , int64_t   open_count , int64_t   read_count , int64_t   write_count , int64_t   seek_count , int64_t   read_bytes , int64_t   write_bytes)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_report_file_info_new;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(open_count) );
	assert( syscall_sock->code(read_count) );
	assert( syscall_sock->code(write_count) );
	assert( syscall_sock->code(seek_count) );
	assert( syscall_sock->code(read_bytes) );
	assert( syscall_sock->code(write_bytes) );
	assert( syscall_sock->code(name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_std_file_info(int   fd , char * &  logical_name)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_std_file_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(logical_name) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_file_info_new(char *  logical_name , char * &  actual_url)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_file_info_new;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(logical_name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(actual_url) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_file_info(char *  logical_name , char * &  actual_url)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_file_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(logical_name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(actual_url) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_std_file_info(int   fd , char * &  logical_name , int *  pipe_fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_std_file_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(pipe_fd) );
	assert( syscall_sock->code(logical_name) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_file_info(char *  logical_name , int *  fd , char * &  physical_name)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_file_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(logical_name) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(physical_name) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_buffer_info(int *  bytes , int *  block_size , int *  prefetch_size)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_buffer_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(prefetch_size) );
	assert( syscall_sock->code(block_size) );
	assert( syscall_sock->code(bytes) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_job_attr(char *  attrname , char * &  expr)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_job_attr;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(attrname) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(expr) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_set_job_attr(char *  attrname , char *  expr)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_set_job_attr;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(expr) );
	assert( syscall_sock->code(attrname) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_constrain(char *  expr)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_constrain;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(expr) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_iwd(char * &  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_iwd;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_ckpt_name(char * &  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_ckpt_name;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_ckpt_mode(condor_signal_t   sig)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_ckpt_mode;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(sig) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_ckpt_speed()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_ckpt_speed;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_a_out_name(char * &  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_a_out_name;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(path) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_pvm_info(char *  response , int   mytid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_pvm_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(mytid) );
	assert( syscall_sock->code(response) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_pvm_task_info(int   v_pid , int   pvm_tid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_pvm_task_info;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(v_pid) );
	assert( syscall_sock->code(pvm_tid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_suspended(int   susp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_suspended;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(susp) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_get_file_stream(char *  file , size_t *  len , unsigned int *  ip_addr , u_short *  port_num)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_get_file_stream;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(file) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(port_num) );
	assert( syscall_sock->code(ip_addr) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_put_file_stream(char *  file , size_t   len , unsigned int *  ip_addr , u_short *  port_num)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_put_file_stream;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code(file) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->code(port_num) );
	assert( syscall_sock->code(ip_addr) );
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_ckpt_server(char *  host)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_ckpt_server;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(host) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_arch(char *  arch)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_arch;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(arch) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_opsys(char *  opsys)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_opsys;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(opsys) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

int  
REMOTE_CONDOR_register_begin_execution()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_register_begin_execution;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_mprotect )
int  
REMOTE_CONDOR_mprotect(MMAP_T   addr , size_t   len , int   prot)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_mprotect;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(addr) );
	assert( syscall_sock->code(len) );
	assert( syscall_sock->code(prot) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setuid )
int  
REMOTE_CONDOR_setuid(uid_t   uid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setuid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(uid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setgid )
int  
REMOTE_CONDOR_setgid(gid_t   gid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setgid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(gid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setreuid )
int  
REMOTE_CONDOR_setreuid(uid_t   ruid , uid_t   euid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setreuid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(ruid) );
	assert( syscall_sock->code(euid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setregid )
int  
REMOTE_CONDOR_setregid(uid_t   rgid , uid_t   egid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setregid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(rgid) );
	assert( syscall_sock->code(egid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setlogin )
int  
REMOTE_CONDOR_setlogin(char *  name)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setlogin;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_revoke )
int  
REMOTE_CONDOR_revoke(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_revoke;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_chroot )
int  
REMOTE_CONDOR_chroot(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_chroot;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setgroups )
int  
REMOTE_CONDOR_setgroups(int   gidsetsize , gid_t  *  grouplist)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setgroups;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(gidsetsize) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_plock )
int  
REMOTE_CONDOR_plock(int   op)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_plock;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(op) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_adjtime )
int  
REMOTE_CONDOR_adjtime(struct timeval *  delta , struct timeval *  old_delta)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_adjtime;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_sethostid )
int  
REMOTE_CONDOR_sethostid(int   host_id)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_sethostid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(host_id) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_async_daemon )
int  
REMOTE_CONDOR_async_daemon()
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_async_daemon;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getfh )
int  
REMOTE_CONDOR_getfh(int   fd , fhandle_t *  fhp , int   exp_fd)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getfh;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(fd) );
	assert( syscall_sock->code(exp_fd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_exportfs )
int  
REMOTE_CONDOR_exportfs(int   access , int *  record , struct exportfsdata *  exdata)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_exportfs;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(access) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_mount )
int  
REMOTE_CONDOR_mount(char *  spec , char *  dir , char *  fstyp , unsigned long   rwflag , void *  data)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_mount;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(rwflag) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_umount )
int  
REMOTE_CONDOR_umount(char *  dir)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_umount;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_acct )
int  
REMOTE_CONDOR_acct(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_acct;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_reboot )
int  
REMOTE_CONDOR_reboot(int   magic , int   magic_too , int   flag)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_reboot;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(magic) );
	assert( syscall_sock->code(magic_too) );
	assert( syscall_sock->code(flag) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_sethostname )
int  
REMOTE_CONDOR_sethostname(char *  address , size_t   address_len)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_sethostname;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(address_len) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_settimeofday )
int  
REMOTE_CONDOR_settimeofday(struct timeval *  tp , struct timezone *  tzp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_settimeofday;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setdomainname )
int  
REMOTE_CONDOR_setdomainname(char *  name , size_t   namelen)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setdomainname;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(namelen) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_swapon )
int  
REMOTE_CONDOR_swapon(char *  path , int   swapflags)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_swapon;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(swapflags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_swapoff )
int  
REMOTE_CONDOR_swapoff(char *  path)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_swapoff;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_bdflush )
int  
REMOTE_CONDOR_bdflush(int   func , long   data)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_bdflush;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(func) );
	assert( syscall_sock->code(data) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_create_module )
caddr_t  
REMOTE_CONDOR_create_module(char *  name , size_t   size)
{
	int	scm;
	caddr_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_create_module;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(size) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val == NULL  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_delete_module )
caddr_t  
REMOTE_CONDOR_delete_module(char *  name)
{
	int	scm;
	caddr_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_delete_module;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val == NULL  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_init_module )
int  
REMOTE_CONDOR_init_module(char *  name , struct module *  image)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_init_module;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_iopl )
int  
REMOTE_CONDOR_iopl(int   level)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_iopl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(level) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_idle )
void  
REMOTE_CONDOR_idle()
{
	int	scm;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_idle;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	_condor_signals_enable( omask );
	SetSyscalls( scm );
}

#endif

#if defined( SYS_ioperm )
int  
REMOTE_CONDOR_ioperm(unsigned long   from , unsigned long   num , int   turn_on)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_ioperm;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(from) );
	assert( syscall_sock->code(num) );
	assert( syscall_sock->code(turn_on) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

pid_t  
REMOTE_CONDOR_fork()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_fork;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#if defined( SYS_vfork )
pid_t  
REMOTE_CONDOR_vfork()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_vfork;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_execve )
int  
REMOTE_CONDOR_execve(char *  path , char * *  argv , char * *  envp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_execve;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_execv )
int  
REMOTE_CONDOR_execv(char *  path , char * *  argv)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_execv;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_exec_with_loader )
int  
REMOTE_CONDOR_exec_with_loader(int   flags , char *  loader , char *  file , char * *  argv , char * *  envp)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_exec_with_loader;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_execvp )
int  
REMOTE_CONDOR_execvp(char *  file , char * *  argv)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_execvp;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_wait4 )
pid_t  
REMOTE_CONDOR_wait4(pid_t   pid , void *  status , int   options , struct rusage *  ru)
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_wait4;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(pid) );
	assert( syscall_sock->code(options) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_wait )
pid_t  
REMOTE_CONDOR_wait(int *  status)
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_wait;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_waitpid )
pid_t  
REMOTE_CONDOR_waitpid(pid_t   pid , int *  status , int   options)
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_waitpid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(pid) );
	assert( syscall_sock->code(options) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_ptrace )
int  
REMOTE_CONDOR_ptrace(long   req , long int   process , unsigned long *  address , unsigned long   data)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_ptrace;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(req) );
	assert( syscall_sock->code(process) );
	assert( syscall_sock->code(data) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setpgid )
int  
REMOTE_CONDOR_setpgid(pid_t   process_id , pid_t   gid)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setpgid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(process_id) );
	assert( syscall_sock->code(gid) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setsid )
pid_t  
REMOTE_CONDOR_setsid()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setsid;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setpgrp )
pid_t  
REMOTE_CONDOR_setpgrp()
{
	int	scm;
	pid_t     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setpgrp;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_profil )
int  
REMOTE_CONDOR_profil(char *  buf , int   buf_size , int   offset , int   scale)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_profil;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(buf_size) );
	assert( syscall_sock->code(offset) );
	assert( syscall_sock->code(scale) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_quotactl )
int  
REMOTE_CONDOR_quotactl(char *  path , int   cmd , int   id , char *  addr)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_quotactl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->code(id) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getsysinfo )
int  
REMOTE_CONDOR_getsysinfo(unsigned long   op , char *  buf , unsigned long   nbytes , int *  start , char *  arg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getsysinfo;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(op) );
	assert( syscall_sock->code(nbytes) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_setitimer )
int  
REMOTE_CONDOR_setitimer(int   which , struct itimerval *  value , struct itimerval *  oval)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_setitimer;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(which) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_getitimer )
int  
REMOTE_CONDOR_getitimer(int   which , struct itimerval *  value)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_getitimer;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(which) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_msgctl )
int  
REMOTE_CONDOR_msgctl(int   msgid , int   cmd , struct msqid_ds *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_msgctl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(msgid) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_msgget )
int  
REMOTE_CONDOR_msgget(key_t   key , int   msgflg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_msgget;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(key) );
	assert( syscall_sock->code(msgflg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_msgrcv )
int  
REMOTE_CONDOR_msgrcv(int   qid , void *  msgp , size_t   msgsz , long   msgtyp , int   msgflg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_msgrcv;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(qid) );
	assert( syscall_sock->code(msgsz) );
	assert( syscall_sock->code(msgtyp) );
	assert( syscall_sock->code(msgflg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_msgsnd )
int  
REMOTE_CONDOR_msgsnd(int   msqid , void *  msgp , size_t   msgsz , int   msgflg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_msgsnd;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(msqid) );
	assert( syscall_sock->code(msgsz) );
	assert( syscall_sock->code(msgflg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_semctl )
int  
REMOTE_CONDOR_semctl(int   semid , int   semnum , int   cmd , void *  arg)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_semctl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(semid) );
	assert( syscall_sock->code(semnum) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_semget )
int  
REMOTE_CONDOR_semget(key_t   key , int   nsems , int   semflag)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_semget;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(key) );
	assert( syscall_sock->code(nsems) );
	assert( syscall_sock->code(semflag) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_semop )
int  
REMOTE_CONDOR_semop(int   semid , struct sembuf *  sops , unsigned int   nsops)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_semop;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(semid) );
	assert( syscall_sock->code(nsops) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_shmctl )
int  
REMOTE_CONDOR_shmctl(int   shmid , int   cmd , struct shmid_ds *  buf)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shmctl;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(shmid) );
	assert( syscall_sock->code(cmd) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_shmdt )
int  
REMOTE_CONDOR_shmdt(void *  addr)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shmdt;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_shmget )
int  
REMOTE_CONDOR_shmget(key_t   key , size_t   size , int   flags)
{
	int	scm;
	int     ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shmget;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(key) );
	assert( syscall_sock->code(size) );
	assert( syscall_sock->code(flags) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val < 0  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

#if defined( SYS_shmat )
void * 
REMOTE_CONDOR_shmat(int   shmid , void *  shmaddr , int   shmflg)
{
	int	scm;
	void *    ret_val;
	condor_errno_t   terrno;
	sigset_t	omask;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();
	CurrentSysCall = CONDOR_shmat;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(shmid) );
	assert( syscall_sock->code(shmflg) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(ret_val) );
	if(  ret_val == NULL  ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return ret_val;
	}
	assert( syscall_sock->end_of_message() );
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return ret_val;
}

#endif

/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

/* Converted case statements to function calls instead. -psilord */

#ifdef LINK_PVM   
	// only include CONDOR_work_request when compiling up the PVM
	// starter.  It is used to send startup job info between the PVM
	// starter and shadow.  we do not want it in the regular condor_syscall_lib
	// because it sends Class Ads, and we do not want to force users to
	// link their STANDARD universe jobs with the classad library. -Todd 12/97
int
REMOTE_CONDOR_work_request(ClassAd *ad, char *&a_out, char *&targ, char *&orig,
	condor_signal_t *kill_sig)
{
	int scm;
	int rval;
	condor_errno_t terrno;
	sigset_t    omask;

	assert( a_out == NULL );
	assert( targ == NULL );
	assert( orig == NULL );

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();

	CurrentSysCall = CONDOR_work_request;

	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->end_of_message() );

	syscall_sock->decode();
	assert( syscall_sock->code(rval) );
	if( rval < 0 ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return rval;
	}
	assert( syscall_sock->code(kill_sig) );
	assert( syscall_sock->code(orig) );
	assert( syscall_sock->code(targ) );
	assert( syscall_sock->code(a_out) );
	assert( getClassAd(syscall_sock, *ad) );
	assert( syscall_sock->end_of_message() );

	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return rval;
}
#endif

int
REMOTE_CONDOR_send_a_file(char *local, char *remote, mode_t mode)
{
	int		loc_fd;
	int		read_status = 0, write_status;
	char	buf[ STD_UNIV_IO_BUF_SIZE ];
	int		len;
	size_t	bytes_to_go;
	int		file_size;
	int scm;
	int rval;
	condor_errno_t terrno;
	sigset_t    omask;


	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();

	CurrentSysCall = CONDOR_send_a_file;

		/* Do the local open */
	if( (loc_fd = safe_open_wrapper(local,O_RDONLY)) < 0 ) {
		rval = -1;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		return rval;
	}

		/* Send over the file name, and mode */
	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(mode) );
	assert( syscall_sock->code(remote) );
	assert( syscall_sock->end_of_message() );

		/* Grab the result of the remote open */
	syscall_sock->decode();
	assert( syscall_sock->code(rval) );
	if( rval < 0 ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		rval = -1;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return rval;
	}
	assert( syscall_sock->end_of_message() );

		/* Send file length */
	file_size = lseek( loc_fd, 0, 2 );
	lseek( loc_fd, 0, 0 );
	syscall_sock->encode();
	assert( syscall_sock->code(file_size) );

		/* Transfer the data */
	for( bytes_to_go = file_size; bytes_to_go; bytes_to_go -= len ) {
		len = bytes_to_go < sizeof(buf) ? bytes_to_go : sizeof(buf);
		read_status = read( loc_fd, buf, len );
		assert( syscall_sock->code_bytes_bool(buf,len) );
	}
	(void)close( loc_fd );

		/* As a check, re-send the length */
	assert( syscall_sock->code(file_size) );
	assert( syscall_sock->end_of_message() );


		/* Get cumulative status from remote writes */
	syscall_sock->decode();
	assert( syscall_sock->code(write_status) );
	if( write_status < 0 ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		rval = write_status;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return rval;
	}
	assert( syscall_sock->end_of_message() );

		/* If we had a local error on the reads, we had to ignore
		   it until now to avoid getting out of sync with the
		   shadow - now we can deal with it.
		*/
	if( read_status < 0 ) {
		rval = read_status;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		return rval;
	}

		/* Everything worked - status is cumulative status from
		   remote writes.
		*/
	rval = write_status;
	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return rval;
}

int
REMOTE_CONDOR_get_file(char *remote, char *local, int mode)
{
	int		loc_fd;
	int		write_status = 0, read_status;
	char	buf[ STD_UNIV_IO_BUF_SIZE ];
	unsigned		checksum;
	size_t		len;
	size_t		file_size;
	size_t		bytes_to_go;
	int		scm;
	int		close_status;
	int		rval;
	sigset_t	omask;
	condor_errno_t	terrno;

	scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
	omask = _condor_signals_disable();

	CurrentSysCall = CONDOR_get_file;

		/* Open local file for writing */
	loc_fd = safe_open_wrapper( local, O_WRONLY|O_CREAT|O_TRUNC, mode );
	if( loc_fd < 0 ) {
		rval = -1;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		return rval;
	}

		/* Send over the file name */
	syscall_sock->encode();
	assert( syscall_sock->code(CurrentSysCall) );
	assert( syscall_sock->code(remote) );
	assert( syscall_sock->end_of_message() );

		/* Grab the result of the remote open */
	syscall_sock->decode();
	assert( syscall_sock->code(rval) );
	if( rval < 0 ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		rval = -1;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return rval;
	}

		/* Get the size of the file */
	assert( syscall_sock->code(file_size) );

		/* Transfer the data */
	for( bytes_to_go = file_size; bytes_to_go; bytes_to_go -= len ) {
		len = bytes_to_go < sizeof(buf) ? bytes_to_go : sizeof(buf);
		assert( syscall_sock->code_bytes_bool(buf,len) );
		write_status = write( loc_fd, buf, len );
	}
	close_status = close( loc_fd );
	if( close_status < 0 ) {
		dprintf( D_ALWAYS,
			"Close of \"%s\" failed - errno = %d\n", local, errno
		);
	}

		/* Should get file size again as a check */
	assert( syscall_sock->code(checksum) );
	assert( checksum == file_size );

		/* Get status from remote reads */
	assert( syscall_sock->code(read_status) );
	if( read_status < 0 ) {
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );
		rval = -1;
		_condor_signals_enable( omask );
		SetSyscalls( scm );
		errno = (int)terrno;
		return rval;
	}
	assert( syscall_sock->end_of_message() );

	/* Wrapup */
	if( write_status >= 0 && read_status >= 0 && close_status >= 0 ) {
		dprintf( D_ALWAYS,
			"CONDOR_get_file: transferred %lu bytes\n", 
			(unsigned long)file_size );
		rval = 0;
	} else {
		rval = -1;
	}

	_condor_signals_enable( omask );
	SetSyscalls( scm );
	return rval;
}

// Stolen from the code generator.
#if defined(LINUX)

#if defined( SYS_mmap )
void * REMOTE_CONDOR_mmap(void * a, size_t l, int p, int f, int fd, off_t o) {
    int scm;
    void *    ret_val;
    condor_errno_t   terrno;
    sigset_t    omask;

    scm = SetSyscalls( SYS_LOCAL | SYS_UNMAPPED );
    omask = _condor_signals_disable();
    CurrentSysCall = CONDOR_mmap;

    syscall_sock->encode();
    assert( syscall_sock->code(CurrentSysCall) );
    assert( syscall_sock->code(l) );
    assert( syscall_sock->code(p) );
    assert( syscall_sock->code(f) );
    assert( syscall_sock->code(fd) );
    assert( syscall_sock->code(o) );
    assert( syscall_sock->end_of_message() );

    syscall_sock->decode();
    assert( syscall_sock->code(ret_val) );
    if(  ret_val == MAP_FAILED  ) {
        assert( syscall_sock->code(terrno) );
        assert( syscall_sock->end_of_message() );
        _condor_signals_enable( omask );
        SetSyscalls( scm );
        errno = (int)terrno;
        return ret_val;
    }
    assert( syscall_sock->end_of_message() );
    _condor_signals_enable( omask );
    SetSyscalls( scm );
    return ret_val;
}
#endif // defined( SYS_mmap )

#endif // defined( LINUX )

} /* end of extern "C" from prologue */
